# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/07_augment_kornia.ipynb (unless otherwise specified).

__all__ = ['KorniaBase']

# Cell
from fastai.vision.all import *
import PIL
import kornia as K
import torchvision.transforms.functional as TTF
from PIL import Image

# Cell
class KorniaBase(RandTransform):
    '''
    Pass in a kornia function, module, list of modules, or nn.Sequential
    containers to `kornia_tfm`.
    If passing functions, you can pass in function arguments as keyword
    args (**kwargs), which can also be random number generators.

    Example
    =======
    * KorniaWrapper(kornia.adjust_hue, hue_factor=1.2)
    * KorniaWrapper(kornia.augmentation.ColorJitter(.2,.3,.1,.2))
    * KorniaWrapper(nn.Sequential(*[kornia.augmentation.ColorJitter()]))
    * KorniaWrapper([
        kornia.augmentation.ColorJitter(.2),
        kornia.augmentation.RandomMotionBlur(3, 5., 1.)
    ]))
    '''
    order = 10
    split_idx = 0
    def __init__(self, kornia_tfm=None,p=1., **kwargs):
        super().__init__(p=p)
        self.tfm = kornia_tfm
        self.input_kwargs = kwargs
        self.call_kwargs  = dict.fromkeys(kwargs)
        self._pipe = Pipeline([ToTensor(), IntToFloatTensor()])
        self.process_tfm()

    def before_call(self, b, split_idx, verbose=False):
        'Compute `p` of applying transform, process input kwargs if applicable'
        self.do = self.p==1. or random.random() < self.p
        for arg,value in self.input_kwargs.items():
            if hasattr(value, '__call__'): self.call_kwargs[arg] = value()
            else: self.call_kwargs[arg] = value

    def process_tfm(self):
        'Process the input `kornia_tfm` argument and make it callable'
        if hasattr(self.tfm, 'forward') and hasattr(self.tfm, '__iter__'):
            pass                                ## -- nn.Sequential

        elif hasattr(self.tfm, 'forward') and type(self.tfm) is not type:
            self.tfm = nn.Sequential(self.tfm)  ## -- Kornia module (called)

        elif hasattr(self.tfm, 'forward') and type(self.tfm) is type:
            #self.tfm = nn.Sequential(self.tfm)  ## -- Kornia module (uncalled)
            pass

        elif isinstance(self.tfm, list):
            self.tfm = nn.Sequential(*self.tfm) ## -- list of Kornia Modules

    def _encode(self, o:TensorImage): return TensorImage(self.tfm(o, **self.call_kwargs)) if self.do else o
    def encodes(self, o:torch.Tensor): return self._encode(o)
    def encodes(self, o:Image.Image):  return self._encode(self._pipe(PILImage(o)))
    def encodes(self, o:TensorImage):  return self._encode(o)
    def encodes(self, o:PILImage):     return self._encode(self._pipe(o))
    def encodes(self, o:(str,Path)):   return self._encode(self._pipe(PILImage.create(o)))
    def encodes(self, o:(TensorCategory,TensorMultiCategory)): return o

    def __repr__(self): return self.tfm.__repr__()


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/03_interpret_gradcam.ipynb (unless otherwise specified).

__all__ = ['Hook', 'HookBwd', 'create_test_img', 'compute_gcam_items', 'compute_gcam_map', 'plt_decoded', 'plot_gcam',
           'GradCam', 'PathLike']

# Cell
try:
    from fastai.vision.all import *
except:
    from fastai2.vision.all import *
from typing import List, Tuple, Callable, Union, Optional, Any

# Cell
class Hook():
    def __init__(self, m):
        self.hook = m.register_forward_hook(self.hook_func)
    def hook_func(self, m, inp, out):  self.stored = out.detach().clone()
    def __enter__(self, *args): return self
    def __exit__ (self, *args): self.hook.remove()

class HookBwd():
    def __init__(self,m):
        self.hook = m.register_backward_hook(self.hook_func)
    def hook_func(self, model, grad_in, grad_out): self.stored = grad_out[0].detach().clone()
    def __enter__(self, *args): return self
    def __exit__ (self, *args): self.hook.remove()

# Cell
def create_test_img(learn, f, return_img=True):
    img = PILImage.create(f)
    x = first(learn.dls.test_dl([f]))
    x = x[0]
    if return_img: return img,x
    return x

# Cell
def compute_gcam_items(learn, x, label) -> Tuple[torch.Tensor]:
    'Compute gradient and activations of `model` for `x` with respect to `label`'

    label_idx = learn.dls.vocab.o2i[label]

    with HookBwd(learn.model[0]) as hook_g:
        with Hook(learn.model[0]) as hook:
            preds       = learn.model.eval()(x)
            activations = hook.stored
        preds[0, label_idx].backward()
        gradients = hook_g.stored

    preds = getattr(learn.loss_func, 'activation', noop)(preds)

    # remove leading batch_size axis
    gradients   = gradients  [0]
    activations = activations[0]

    return gradients,activations,preds.detach().cpu().numpy().flatten()

# Cell
def compute_gcam_map(gradients, activations):
    '(mean(gradients) * activations).sum() to return tiny grad-cam map'
    # Mean over the feature maps. If you don't use `keepdim`, it returns
    # a value of shape (1280) which isn't amenable to `*` with the activations
    gcam_weights = gradients.mean(dim=[1,2], keepdim=True) # (1280,7,7)   --> (1280,1,1)
    gcam_map     = (gcam_weights * activations) # (1280,1,1) * (1280,7,7) --> (1280,7,7)
    gcam_map     = gcam_map.sum(0)              # (1280,7,7) --> (7,7)
    return gcam_map

# Cell
def plt_decoded(learn, x, ctx):
    'Processed tensor --> plottable image, return `extent`'
    x_decoded = TensorImage(learn.dls.train.decode((x,))[0][0])
    extent = (0, x_decoded.shape[1], x_decoded.shape[2], 0)
    x_decoded.show(ctx=ctx)
    return extent

def plot_gcam(img:PILImage, x:tensor, gcam_map:tensor, plt_axis,
              full_size=True, alpha=0.6, learn=None,
              interpolation='bilinear', cmap='magma'):
    'Plot gradcam on `plt_axis`'
    if full_size:
        extent = (0, img.width,img.height, 0)
        show_image(img, ctx=plt_axis)
    else:
        extent = plt_decoded(learn, x, plt_axis)

    show_image(gcam_map.detach().cpu(), ctx=plt_axis,
               alpha=alpha, extent=extent,
               interpolation=interpolation, cmap=cmap)

# Cell
import math
from typing import List
PathLike = Union[str,Path]

class GradCam():
    "Class interface to facilitate computing and siplaying Grad-CAM"
    def __init__(self, learn:Learner, fname:PathLike, labels:Union[str,List[str],None]):
        """
        Compute Grad-CAM maps for all `labels`.
        If `labels` is None, compute for the predicted class (more expensive)
        """
        self.learn = learn
        self.fname = fname
        self.img, self.x = create_test_img(self.learn, self.fname)
        if labels is None:
            self.labels = [self.learn.predict(fname)[0]]
        else:
            self.labels = [labels] if isinstance(labels,str) else labels
        self.compute_gcams()

    def compute_gcams(self):
        self.gradcams = defaultdict()
        for label in self.labels:
            gradients, activations, self.preds = compute_gcam_items(self.learn,self.x,label)
            gcam_map = compute_gcam_map(gradients, activations)
            self.gradcams[label] = gcam_map
            self.preds_dict = {
                lab:pred for pred,lab in zip(self.preds, self.learn.dls.vocab)
            }

    def plot(self, max_ncols=None, full_size=True, alpha=0.6,
             interpolation='bilinear', cmap='magma',
             figsize=(12,12), return_fig=False, plot_original=False):
        """
        Plot the computed Grad-CAMs.

        Key Arguments
        -------------
        * full_size: If True, plots the images in their original size, else
                     in the size that the `Learner` resizes them to
        * plot_original: if True, plots the original image without any overlays
                         in addition to the heatmaps
        * max_ncols: Use this to manipulate the number of rows you'd like your
                     plot to have. Useful for classifiers with a large no. of
                     outputs. Enter `None` to plot everything in one row.
        """
        label_idx = 0
        total = len(self.labels)+1 if plot_original else len(self.labels)

        if max_ncols is None:
            max_ncols=len(self.labels)+1 if plot_original else len(self.labels)

        if total > max_ncols:
            nrows  = math.ceil(total/max_ncols)
            fig,ax = plt.subplots(nrows=nrows, ncols=max_ncols, figsize=figsize)
            plt.axis('off')

            for i in range(nrows):
                for j in range(max_ncols):
                    if plot_original:
                        if i==0 and j==0:
                            if full_size:
                                show_image(self.img, ctx=ax[0,0])
                            else:
                                x = TensorImage(learn.dls.train.decode((self.x,))[0][0])
                                x.show(ctx=ax[0,0])
                            ax[0][0].set_title('original')
                            continue
                    plot_gcam(img=self.img, x=self.x, full_size=full_size,
                              gcam_map=self.gradcams[self.labels[label_idx]],
                              plt_axis=ax[i,j], alpha=alpha, learn=self.learn,
                              interpolation=interpolation, cmap=cmap)
                    title = self.labels[label_idx]
                    ax[i][j].set_title(f'{title}, {self.preds_dict[title] * 100:.02f}%')
                    label_idx += 1
                    if label_idx >= len(self.labels): break
        else:
            fig,ax = plt.subplots(nrows=1, ncols=max_ncols, figsize=figsize)
            plt.axis('off')

            for i in range(len(ax)):
                if plot_original:
                    if i==0 and full_size:
                        show_image(self.img, ctx=ax[0])
                        ax[0].set_title('original')
                        continue
                    elif i==0 and not full_size:
                        _ = plt_decoded(self.learn, self.x, ctx=ax[0])
                        ax[0].set_title('original')
                        continue
                plot_gcam(img=self.img, x=self.x, full_size=full_size,
                          gcam_map=self.gradcams[self.labels[label_idx]],
                          plt_axis=ax[i], alpha=alpha, learn=self.learn,
                          interpolation=interpolation, cmap=cmap)
                title = self.labels[label_idx]
                ax[i].set_title(f'{title}, {self.preds_dict[title] * 100:.02f}%')
                label_idx += 1


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/06_augment_albumentations.ipynb (unless otherwise specified).

__all__ = ['albu_augment', 'AlbumentationsWrapper', 'BlurringTfms', 'StyleTfms', 'WeatherTfms', 'NoiseTfms',
           'ColorTonesTfms', 'ColorChannelTfms', 'LightingTfms', 'OtherTfms', 'Tfms']

# Cell
try:
    from fastai.vision.all import *
except:
    from fastai2.vision.all import *
import PIL
from typing import List, Tuple, Callable, Union, Optional, Any

# Cell
import albumentations as A
from albumentations import (
    RandomBrightness, RandomContrast, CLAHE,      # Lighting Transforms
    ToSepia, ToGray,                              # Color Tones
    RGBShift, HueSaturationValue, ChannelShuffle, # Color Channels
    Blur, MotionBlur, GaussianBlur,               # Blurs
    MedianBlur, IAAEmboss,                        # Style Transfer-Esque
    GaussNoise, IAAAdditiveGaussianNoise,         # Noise
    JpegCompression, Posterize,                   # Noise
    RandomSunFlare,                               # Weather
    FancyPCA,                                     # Other
    OneOf, Compose
)

# Cell
def albu_augment(aug, img): return aug(image=np.array(img))['image']

##https://dev.fast.ai/tutorial.pets
class AlbumentationsWrapper(Transform):
    "Wrapper function for any Albumentations Transform"
    order=0 # apply before `ToTensor`
    def __init__(self, aug): self.aug=aug
    def __repr__(self): return self.aug.__repr__()

    def _encodes(self, img:PILImage):
        aug_img = albu_augment(self.aug, img)
        return PILImage.create(aug_img)

    def encodes(self, img:PIL.Image.Image): return self._encodes(img)
    def encodes(self, img:PILImage): return self._encodes(img)
    def encodes(self, o:(str,Path)): return self._encodes(PILImage.create(o))

# Cell

# lean heavily towards Motion Blur
BlurringTfms = OneOf([
    MotionBlur(blur_limit=(15,15), p=0.8),
    Blur(p=0.2)
], p=0.5)

StyleTfms = OneOf([
    MedianBlur(blur_limit=(3,7), p=0.6),
    IAAEmboss(strength=(0.2,0.99), p=0.4)
], p=0.3)

WeatherTfms = RandomSunFlare(
    src_radius=80, p=0.1
)

NoiseTfms = OneOf([
    GaussNoise(p=0.6),
    IAAAdditiveGaussianNoise(p=0.4), # stronger
    JpegCompression(quality_lower=25, quality_upper=55, p=0.2)
], p=0.25)

ColorTonesTfms = OneOf([
    ToSepia(),
    ToGray()
], p=0.3)

ColorChannelTfms = OneOf([
    ChannelShuffle(),
    HueSaturationValue(val_shift_limit=5),
    RGBShift()
], p=0.3)

LightingTfms = OneOf([
    RandomContrast(p=0.1),
    RandomBrightness(p=0.1),
    CLAHE(p=0.8)
], p=0.3)

OtherTfms = FancyPCA(alpha=0.4, p=0.4)

# Cell
Tfms = Compose([
    BlurringTfms,
    StyleTfms,
    WeatherTfms,
    NoiseTfms,
    ColorTonesTfms,
    ColorChannelTfms,
    LightingTfms,
    OtherTfms
])


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/04_inference_export-to-other-frameworks.ipynb (unless otherwise specified).

__all__ = ['to_cuda', 'preprocess_one', 'preprocess_batch', 'PathLike', 'open_image', 'torch_to_onnx', 'torch_to_numpy',
           'onnx_to_tf', 'get_nn_spec', 'onnx_to_coreml']

# Cell
try:
    from fastai.vision.all import *
except:
    from fastai2.vision.all import *
from typing import List, Tuple, Callable, Union, Optional, Any

# Cell
import torchvision.transforms.functional as TTF
import torch.nn as nn
to_cuda = lambda x: x.cuda() if torch.cuda.is_available() else x

# Cell
from typing import Union,Collection

PathLike = Union[str,Path]

def preprocess_one(fname:PathLike):
    x = open_image(fname)
    x = TTF.to_tensor(x)
    x = TTF.normalize(x, mean=[0.485, 0.456, 0.406], # imagenet mean
                         std =[0.229, 0.224, 0.225]) # imagenet sdev
    x = to_cuda(x)
    x = x.unsqueeze(0)
    return x

def preprocess_batch(fnames:Union[PathLike,Collection]):
    batch = [preprocess_one(f) for f in fnames]
    return torch.cat(batch)

# Cell
from pprint import pprint
import PIL

open_image = lambda f,size=(224,224): PIL.Image.open(f).convert('RGB').resize(size, PIL.Image.BILINEAR)

# Cell
import onnx
import onnx.utils
from onnx import optimizer

# Cell
def torch_to_onnx(model:nn.Module,
                  activation:nn.Module=None,
                  save_path:str     = '../exported-models/',
                  model_fname:str   = 'onnx-model',
                  input_shape:tuple = (1,3,224,224),
                  input_name:str    = 'input_image',
                  output_names:Union[str,list] = 'output',
                  **export_args) -> None:
    """
    Export a `nn.Module` -> ONNX

    This function exports the model with support for batching,
    checks that the export was done properly, and polishes the
    model up (removes unnecessary fluff added during conversion)

    Key Arguments
    =============
    * activation:  If not None, append this to the end of your model.
                   Typically a `nn.Softmax(-1)` or `nn.Sigmoid()`
    * input_shape: Shape of the inputs to the model
    """
    save_path = Path(save_path)
    if isinstance(output_names, str): output_names = [output_names]
    if activation: model = nn.Sequential(*[model, activation])
    model.eval()
    x = torch.randn(input_shape, requires_grad=True)
    x = x.cuda() if torch.cuda.is_available() else x
    model(x)
    dynamic_batch = {0: 'batch'}
    dynamic_axes  = {input_name : dynamic_batch}
    for out in output_names: dynamic_axes[out] = dynamic_batch
    torch.onnx._export(model, x, f"{save_path/model_fname}.onnx",
                       export_params=True, verbose=False,
                       input_names=[input_name], output_names=output_names,
                       dynamic_axes=dynamic_axes, keep_initializers_as_inputs=True,
                       **export_args)
    print(f"Loading, polishing, and optimising exported model from {save_path/model_fname}.onnx")
    onnx_model = onnx.load(f'{save_path/model_fname}.onnx')
    model = onnx.utils.polish_model(onnx_model)
    #onnx.checker.check_model(model)

    # removing unused parts of the model
    passes = ["extract_constant_to_initializer", "eliminate_unused_initializer"]
    optimized_model = optimizer.optimize(onnx_model, passes)

    onnx.save(optimized_model, f'{save_path/model_fname}.onnx')
    print('Exported successfully')

# Cell
def torch_to_numpy(x:tensor):
    if x.requires_grad: return x.detach().cpu().numpy()
    else:               return x.cpu().numpy()

# Cell
import onnx_tf
from onnx_tf.backend import prepare

# Cell
def onnx_to_tf(onnx_model:PathLike, output_path:PathLike):
    """
    Simplest wrapper around https://github.com/onnx/onnx-tensorflow/blob/master/example/onnx_to_tf.py
    Ensure `output_path` includes `.pb` as the file extension
    """
    onnx_model = onnx.load(onnx_model)
    tf_rep     = prepare(onnx_model)  # prepare tf representation
    tf_rep.export_graph(output_path)  # export the model
    return tf_rep

# Cell
import copy
import coremltools
import onnx_coreml
import os

from onnx_coreml import convert

# Cell
# borrowed from the CoreML Survival Guide, written by Matthijs Hollemans
def get_nn_spec(spec):
    "spec is of type `Model_pb2.Model`, accessed via coreml_model.get_spec()"
    if   spec.WhichOneof("Type") == 'neuralNetwork': return spec.neuralNetwork
    elif spec.WhichOneof("Type") == 'neuralNetworkClassifier': return spec.neuralNetworkClassifier
    elif spec.WhichOneof("Type") == 'neuralNetworkRegressor':  return spec.neuralNetworkRegressor
    return None

# Cell
def onnx_to_coreml(onnx_path:PathLike,
                   normalise_mean:Collection=[0.485, 0.456, 0.406],
                   normalise_sdev:Collection=[0.229, 0.224, 0.225],
                   is_bgr:bool=False,
                   mode:Optional[str]='classifier',
                   class_labels:Optional[Collection]=None,
                   image_input_names = ['input_image'],
                   minimum_ios_deployment_target = '11.2',
                   output_path=''
                  ):
    """
    Add a custom scaling layer with the `normalise_mean` and
    `normalise_sdev` stats, do the appropriate preprocessing
    and export an onnx-model to CoreML.

    Ensure that the `output_path` includes the `.mlmodel` extension
    """
    assert '.mlmodel' in output_path, f"Include '.mlmodel' file extension in `output_path`"
    # preprocessing arguments
    args = dict(
        image_scale = 1/255.,
        red_bias    = -normalise_mean[0],
        green_bias  = -normalise_mean[1],
        blue_bias   = -normalise_mean[2],
        is_bgr      = is_bgr
    )
    red_sdev, green_sdev, blue_sdev = normalise_sdev

    # convert onnx model to coreML
    coreml_model = convert(
        model = onnx_path,
        mode  = mode,
        class_labels = class_labels,
        image_input_names  = image_input_names,
        preprocessing_args = args,
        minimum_ios_deployment_target = minimum_ios_deployment_target
    )

    # get network spec from model
    spec = coreml_model.get_spec()
    nn   = get_nn_spec(spec)

    # store away old layers to add back to the reconstructed network
    old_layers = copy.deepcopy(nn.layers)
    del nn.layers[:]

    # names of inputs and outputs of the scaling layer
    input_name  = old_layers[0].input[0]
    output_name = f"{input_name}_scaled"

    # create and add scaling layer to new network
    scale_layer = nn.layers.add()
    scale_layer.name = "scale_layer"
    scale_layer.input.append(input_name)
    scale_layer.output.append(output_name)

    scale_layer.scale.scale.floatValue.extend([
        1/red_sdev, 1/green_sdev, 1/blue_sdev
    ])
    scale_layer.scale.shapeScale.extend([3,1,1])

    # add back all the old layers
    nn.layers.extend(old_layers)
    nn.layers[1].input[0] = output_name

    coremltools.utils.save_spec(spec, output_path)
    print(f"Saved CoreML model to {output_path}")
    return coremltools.models.MLModel(output_path)


# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"ClassificationInterpretationEx": "01_interpret_classification-interpretation.ipynb",
         "intersection": "02_interpret_compare-models.ipynb",
         "compare_venn": "02_interpret_compare-models.ipynb",
         "Hook": "03_interpret_gradcam.ipynb",
         "HookBwd": "03_interpret_gradcam.ipynb",
         "create_test_img": "03_interpret_gradcam.ipynb",
         "compute_gcam_items": "03_interpret_gradcam.ipynb",
         "compute_gcam_map": "03_interpret_gradcam.ipynb",
         "plt_decoded": "03_interpret_gradcam.ipynb",
         "plot_gcam": "03_interpret_gradcam.ipynb",
         "GradCam": "03_interpret_gradcam.ipynb",
         "PathLike": "04_inference_export-to-other-frameworks.ipynb",
         "to_cuda": "04_inference_export-to-other-frameworks.ipynb",
         "preprocess_one": "04_inference_export-to-other-frameworks.ipynb",
         "preprocess_batch": "04_inference_export-to-other-frameworks.ipynb",
         "open_image": "04_inference_export-to-other-frameworks.ipynb",
         "torch_to_onnx": "04_inference_export-to-other-frameworks.ipynb",
         "torch_to_numpy": "04_inference_export-to-other-frameworks.ipynb",
         "onnx_to_tf": "04_inference_export-to-other-frameworks.ipynb",
         "get_nn_spec": "04_inference_export-to-other-frameworks.ipynb",
         "onnx_to_coreml": "04_inference_export-to-other-frameworks.ipynb",
         "is_3dlut_row": "05_augment_PIL-img_filters.ipynb",
         "read_lut": "05_augment_PIL-img_filters.ipynb",
         "ApplyPILFilter": "05_augment_PIL-img_filters.ipynb",
         "albu_augment": "06_augment_albumentations.ipynb",
         "AlbumentationsWrapper": "06_augment_albumentations.ipynb",
         "BlurringTfms": "06_augment_albumentations.ipynb",
         "StyleTfms": "06_augment_albumentations.ipynb",
         "WeatherTfms": "06_augment_albumentations.ipynb",
         "NoiseTfms": "06_augment_albumentations.ipynb",
         "ColorTonesTfms": "06_augment_albumentations.ipynb",
         "ColorChannelTfms": "06_augment_albumentations.ipynb",
         "LightingTfms": "06_augment_albumentations.ipynb",
         "OtherTfms": "06_augment_albumentations.ipynb",
         "Tfms": "06_augment_albumentations.ipynb",
         "KorniaBase": "07_augment_kornia.ipynb"}

modules = ["interpret/interpret.py",
           "interpret/compare.py",
           "interpret/gradcam.py",
           "inference/export.py",
           "augment/pil_filters.py",
           "augment/albumentations.py",
           "augment/kornia.py"]

doc_url = "https://Synopsis.github.io/fastai2_extensions/"

git_url = "https://github.com/Synopsis/fastai2_extensions/tree/master/"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/01_interpret_classification-interpretation.ipynb (unless otherwise specified).

__all__ = ['ClassificationInterpretationEx']

# Cell
try:
    from fastai.vision.all import *
    from fastai.metrics import *
except:
    from fastai2.vision.all import *
    from fastai2.metrics import *

# Cell
class ClassificationInterpretationEx(ClassificationInterpretation):
    """
    Extend fastai2's `ClassificationInterpretation` to analyse model predictions in more depth
    """
    def plot_confusion_matrix(interp, normalize=False, title='Confusion matrix', cmap="Blues", norm_dec=2,
                              plot_txt=True, return_fig=False, **kwargs):
        """
        Plot the confusion matrix, with `title` and using `cmap`.
        An exact replica of fastai2's method, with the added option
        of `return_fig`, to be able to save the image to disk
        """
        # This function is mainly copied from the sklearn docs
        cm = interp.confusion_matrix()
        if normalize: cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]
        fig = plt.figure(**kwargs)
        plt.imshow(cm, interpolation='nearest', cmap=cmap)
        plt.title(title)
        tick_marks = np.arange(len(interp.vocab))
        plt.xticks(tick_marks, interp.vocab, rotation=90)
        plt.yticks(tick_marks, interp.vocab, rotation=0)

        if plot_txt:
            thresh = cm.max() / 2.
            for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
                coeff = f'{cm[i, j]:.{norm_dec}f}' if normalize else f'{cm[i, j]}'
                plt.text(j, i, coeff, horizontalalignment="center", verticalalignment="center", color="white" if cm[i, j] > thresh else "black")

        ax = fig.gca()
        ax.set_ylim(len(interp.vocab)-.5,-.5)

        plt.tight_layout()
        plt.ylabel('Actual')
        plt.xlabel('Predicted')
        plt.grid(False)
        if return_fig: return fig

    def compute_label_confidence(self):
        """
        Collate prediction confidence, filenames, and ground truth labels
        in DataFrames, and store them as class attributes
        `self.preds_df` and `self.preds_df_each`
        """
        self._preds_collated = [
            (item, self.dl.vocab[label_idx], *preds.numpy()*100)\
            for item,label_idx,preds in zip(self.dl.items,
                                            self.targs,
                                            self.preds)
        ]

        self.preds_df       = pd.DataFrame(self._preds_collated, columns = ['fname','truth', *self.dl.vocab])
        self._preds_df_each = {l:self.preds_df.copy()[self.preds_df.truth == l].reset_index(drop=True) for l in self.dl.vocab}
        self.preds_df_each  = defaultdict(dict)


        sort_desc = lambda x,col: x.sort_values(col, ascending=False).reset_index(drop=True)
        for label,df in self._preds_df_each.items():
            filt = df[label] == df[self.dl.vocab].max(axis=1)
            self.preds_df_each[label]['accurate']   = df.copy()[filt]
            self.preds_df_each[label]['inaccurate'] = df.copy()[~filt]

            self.preds_df_each[label]['accurate']   = sort_desc(self.preds_df_each[label]['accurate'], label)
            self.preds_df_each[label]['inaccurate'] = sort_desc(self.preds_df_each[label]['inaccurate'], label)
            assert len(self.preds_df_each[label]['accurate']) + len(self.preds_df_each[label]['inaccurate']) == len(df)

    def plot_label_confidence(self, bins:int=10, fig_width:int=12, fig_height_base:int=4,
                              title:str='Accurate vs. Inaccurate Predictions Confidence (%) Levels Per Label',
                              return_fig:bool=False, label_bins:bool=True,
                              accurate_color='mediumseagreen', inaccurate_color='tomato'):
        'Plot label confidence histograms for each label'
        if not hasattr(self, 'preds_df_each'): self.compute_label_confidence()
        fig, axes = plt.subplots(nrows = len(self.preds_df_each.keys()), ncols=2,
                                 figsize = (fig_width, fig_height_base * len(self.dl.vocab)))
        for i, (label, df) in enumerate(self.preds_df_each.items()):
            height=0
            # find max height
            for mode in ['inaccurate', 'accurate']:
                len_bins,_ = np.histogram(df[mode][label], bins=10)
                if len_bins.max() > height: height=len_bins.max()

            for mode,ax in zip(['inaccurate', 'accurate'], axes[i]):
                range_ = (50,100) if mode == 'accurate' else (0,50)
                color  = accurate_color if mode == 'accurate' else inaccurate_color
                num,_,patches = ax.hist(df[mode][label], bins=bins, range=range_, rwidth=.95, color=color)
                if label_bins:
                    for rect in patches:
                        ht = rect.get_height()
                        ax.annotate(s  = f"{int(ht) if ht > 0 else ''}",
                            xy = (rect.get_x() + rect.get_width()/2, ht),
                            xytext = (0,3), # offset vertically by 3 points
                            textcoords = 'offset points',
                            ha = 'center', va = 'bottom'
                           )
                ax.set_ybound(upper=height + height*0.3)
                ax.set_xlabel(f'{label}: {mode.capitalize()}')
                ax.set_ylabel(f'No. {mode.capitalize()} = {len(df[mode][label])}')
        fig.suptitle(title)
        plt.subplots_adjust(top = 0.9, bottom=0.01, hspace=0.25, wspace=0.2)
        if return_fig: return fig


    def plot_accuracy(self, width=0.5, figsize=(6,6), return_fig=False,
                      title='Accuracy Per Label', ylabel='Accuracy (%)',
                      color='steelblue', vertical_labels=False):
        'Plot a bar plot showing accuracy per label'
        if not hasattr(self, 'preds_df_each'): self.compute_label_confidence()
        self.accuracy_dict = defaultdict()

        for label,df in self.preds_df_each.items():
            total = len(df['accurate']) + len(df['inaccurate'])
            self.accuracy_dict[label] = 100 * len(df['accurate']) / total

        fig,ax = plt.subplots(figsize=figsize)

        x = self.accuracy_dict.keys()
        y = [v for k,v in self.accuracy_dict.items()]

        rects = ax.bar(x,y,width,color=color)
        for rect in rects:
            ht = rect.get_height()
            ax.annotate(s  = f"{ht:.02f}",
                        xy = (rect.get_x() + rect.get_width()/2, ht),
                        xytext = (0,3), # offset vertically by 3 points
                        textcoords = 'offset points',
                        ha = 'center', va = 'bottom'
                       )
        ax.set_ybound(lower=0, upper=100)
        ax.set_yticks(np.arange(0,110,10))
        ax.set_ylabel(ylabel)
        ax.set_xticklabels(x, rotation='vertical' if vertical_labels else 'horizontal')
        plt.suptitle(title)
        if return_fig: return fig

    def get_fnames(self, label:str,
                   mode:('accurate','inaccurate'),
                   conf_level:Union[int,float,tuple]) -> np.ndarray:
        """
        Utility function to grab filenames of a particular label `label` that were classified
        as per `mode` (accurate|inaccurate).
        These filenames are filtered by `conf_level` which can be above or below a certain
        threshold (above if `mode` == 'accurate' else below), or in confidence ranges
        """
        assert label in self.dl.vocab
        if not hasattr(self, 'preds_df_each'): self.compute_label_confidence()
        df = self.preds_df_each[label][mode].copy()
        if mode == 'accurate':
            if isinstance(conf_level, tuple):       filt = df[label].between(*conf_level)
            if isinstance(conf_level, (int,float)): filt = df[label] > conf_level
        if mode == 'inaccurate':
            if isinstance(conf_level, tuple):       filt = df[label].between(*conf_level)
            if isinstance(conf_level, (int,float)): filt = df[label] < conf_level
        return df[filt].fname.values


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/02_interpret_compare-models.ipynb (unless otherwise specified).

__all__ = ['intersection', 'compare_venn']

# Cell
try:
    from fastai.vision.all import *
    from fastai.metrics import *
except:
    from fastai2.vision.all import *
    from fastai2.metrics import *
from typing import Collection, Tuple, List, Callable

# Cell
def intersection(sets:Collection):
    '`set.intersection` for a list of sets of any size'
    if len(sets) == 1: return sets
    res = sets[0]
    for x in sets[1:]: res=res.intersection(x)
    return res

# Cell
from matplotlib_venn import venn2, venn3

def compare_venn(interps:Collection,
                 conf_level:Union[int,float,tuple],
                 title=None, mode=('accurate','inaccurate'),
                 return_fig=False, return_common=True,
                 figsize=(10,6), set_color='tomato'):
    """
    Compute the agreement between 2 or more models' predictions

    If you only input 2-3 models in `interps`, then you also get
    a venn diagram to visualise the agreement between these models

    Key Arguments
    =============
    * interps:    a list or tuple of `ClassificationInterpretationEx` objects
    * mode:       either 'accurate' or 'inaccurate' to filter predictions
    * conf_level: a single number or a tuple of (min,max) to filter the prediction confidence
    """
    assert len(interps) > 1, 'Enter 2 or more models to compare'
    ### Extract fnames per model per label
    fnames_dict = defaultdict()
    for i,interp in enumerate(interps):
        name = f"interp{i+1}"
        fnames_dict[name] = {}.fromkeys(interp.dl.vocab)
        for label in fnames_dict[name].keys():
            fnames_dict[name][label] = interp.get_fnames(label, mode, conf_level)

    if len(interps) <= 3:
        ### Prepare sets of filenames per label for venn diagrams
        sets_dict = defaultdict(list)
        for interp in fnames_dict.values():
            for label,fnames in interp.items():
                sets_dict[label].append(set(fnames))

        ### Plot venn diagrams
        fig, axes = plt.subplots(nrows=1, ncols=interps[0].dl.c, figsize=figsize)
        for i,(label,sets) in enumerate(sets_dict.items()):
            set_labels = [f"Model {j+1}" for j in range(len(interps))]
            set_colors = [set_color] * len(interps)
            axes[i].set_title(label)
            if len(interps) == 2:
                venn2(sets, set_labels=set_labels, ax=axes[i], set_colors=set_colors)
            if len(interps) == 3:
                venn3(sets, set_labels=set_labels, ax=axes[i], set_colors=set_colors)

        ## Plot titles and subtitles
        if title is not None: fig.suptitle(title)
        else:
            if isinstance(conf_level, tuple): filler = f'Between {conf_level[0]}-{conf_level[1]}'
            else:
                if mode == 'accurate': filler = f'Above {conf_level}'
                else:                  filler = f'Below {conf_level}'
            if mode == 'accurate':   fig.suptitle(f'Model Agreement - {mode.capitalize()} {filler} % Confidence')
            if mode == 'inaccurate': fig.suptitle(f'Model Agreement - {mode.capitalize()} {filler} % Confidence')
        plt.subplots_adjust(top = 1.1, bottom=0.01, hspace=0.25, wspace=0.1)
    else:
        fig=None

    if return_common:
        res_set = {label:intersection(sets_dict[label]) for label in interps[0].dl.vocab}

    if return_fig    and     return_common: return (fig, res_set)
    if return_fig    and not return_common: return fig
    if return_common and not return_fig:    return res_set


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/05_augment_PIL-img_filters.ipynb (unless otherwise specified).

__all__ = ['is_3dlut_row', 'read_lut', 'ApplyPILFilter']

# Cell
try:
    from fastai.vision.all import *
except:
    from fastai2.vision.all import *
from PIL import ImageFilter
from typing import List, Tuple, Callable, Union, Optional, Any

# Cell
def is_3dlut_row(row:list) -> bool:
    'Check if one line in the file has exactly 3 values'
    row_values = []
    for val in row:
        try: row_values.append(float(val))
        except: continue
    if len(row_values) == 3: return True
    else: return False

def read_lut(path_lut:Union[str,Path], num_channels:int=3):
    'Read LUT from raw file. Assumes each line in a file is part of the lut table'
    with open(path_lut) as f: lut_raw = f.read().splitlines()

    size      = round(len(lut_raw) ** (1/3))
    row2val   = lambda row: tuple([float(val) for val in row])
    lut_table = [row2val(row.split(' ')) for row in lut_raw if is_3dlut_row(row.split(' '))]

    return ImageFilter.Color3DLUT(size, lut_table, num_channels)

# Cell
class ApplyPILFilter(RandTransform):
    "Apply a `PIL.ImageFilter` and return as a PILImage"
    order = 0 # Apply before `ToTensor`
    def __init__(self, filters, p=1.):
        super().__init__(p=p)
        self.filter = filters

    def select_filter(self, o):
        'If multiple `filters` are given, select and apply one'
        if isinstance(self.filter, (tuple,list,L)):
              rand_idx = np.random.randint(0, len(self.filter))
              return o.filter(self.filter[rand_idx])
        else: return o.filter(self.filter)

    #def _encodes(self, o:(PILImage,TensorImage,str,Path)): return TensorImage(self.select_filter(o)).permute(2,0,1)
    def _encodes(self, o): return PILImage(self.select_filter(o))

    def encodes(self, o:PILImage):               return self._encodes(o)
    def encodes(self, o:(TensorImage,str,Path)): return self._encodes(PILImage.create(o))